#include "stdafx.h"
#include "ipccopydata.h"
#include "ipccopydatawindow.h"
#include <algorithm>
#include "shlobj.h"
#include "../checkinexecution.h"

namespace tipbase
{

SHSTDAPI_(BOOL) IsUserAnAdmin(void);
BOOL CheckAdminMode()
{
	return IsUserAnAdmin();
}

IPCStringKey::IPCStringKey( const TCHAR *name )
{
	lstrcpy( m_name, name );
	m_name_size = lstrlen( name ) * 2;
}

void* IPCStringKey::GetData()
{
	return (void *)m_name;
}

int IPCStringKey::GetSize()
{
	return m_name_size;
}

IPCCopyDataServerEventHooker::~IPCCopyDataServerEventHooker()
{
}

int	IPCCopyDataServerEventHooker::Connected( IIPCHandle *connectedHandle )
{
	return 1;
}

void IPCCopyDataServerEventHooker::Close( IIPCHandle *connectedHandle )
{
}

IPCCopyDataServer::IPCCopyDataServer(IPCCopyDataServerEventHooker *hooker)
{
	m_msgWnd = NULL;
	m_hooker = NULL;
	SetHooker(hooker);
}

IPCCopyDataServer::~IPCCopyDataServer()
{
	if( NULL != m_msgWnd )
	{
		delete m_msgWnd;
		m_msgWnd = NULL;
	}
}

//const int MSGFLT_ADD = 1;
//const int MSGFLT_REMOVE = 2;

BOOL VistaMessagePatch()
{
	if( CheckAdminMode() == FALSE || CheckInExecution::GetWindowVersion() < WindowVersion::Type::win_VISTA  )
	{
		return TRUE;
	}

	BOOL bRet = FALSE;

	HMODULE hUser32 = ::LoadLibrary( _T("user32.dll") );
	if( hUser32 != NULL )
	{
		typedef BOOL (WINAPI *LPCHANGEWINDOWMESSAGEFILTER)( UINT message, DWORD dwFlag );
		LPCHANGEWINDOWMESSAGEFILTER lpfn = (LPCHANGEWINDOWMESSAGEFILTER)GetProcAddress( hUser32, "ChangeWindowMessageFilter" );
		if( lpfn != NULL )
		{
			if( lpfn( WM_COPYDATA, /*MSGFLT_ADD*/1 ) )
			{
				bRet = TRUE;
				// 
			}
		}
		::FreeLibrary( hUser32 );
	}

	return bRet;
}

bool IPCCopyDataServer::Create( IIPCKey *host )
{
	VistaMessagePatch();

	if( NULL == host )
		return false;

	m_msgWnd = new IPCCopyDataMessageWindow( this );
	if( NULL == m_msgWnd )
		return false;

	if( NULL == m_msgWnd->Create( HWND_MESSAGE, 0, (LPCTSTR)host->GetData() ) )
	{
		delete m_msgWnd;
		m_msgWnd = NULL;
	}

	lstrcpy( m_hostkey, (LPCTSTR)host->GetData() );
	return true;
}

BOOL IsHandleExist( TCHAR *name )
{
	HWND hWnd = ::FindWindowEx( HWND_MESSAGE, NULL, NULL, name );
	if( NULL == hWnd )
		return FALSE;

	return TRUE;
}

BOOL IPCCopyDataServer::IsExist( IIPCKey *client )
{
	if( NULL == client )
		return FALSE;

	TCHAR *name = (LPTSTR)client->GetData();
	return IsHandleExist( name );
}

IIPCHandle *IPCCopyDataServer::Find( IIPCKey *client, IIPCReceiveCallback *callback )
{
	if( NULL == client )
		return NULL;

	TCHAR *name = (LPTSTR)client->GetData();
	HWND hWnd = ::FindWindowEx( HWND_MESSAGE, NULL, NULL, name );
	if( NULL == hWnd )
		return NULL;

	IIPCHandle *handle = GetHandle( hWnd, name, callback );
	return handle;
}

LRESULT SendIPCCopyDataPacket( HWND targetHwnd, HWND thisHwnd, 
							  IPCCopyDataServer::IPC_COPYDATA_COMMUNICATION_TYPE type, 
								LPVOID sendData, int size)
{
	COPYDATASTRUCT copydata;
	copydata.dwData = (ULONG_PTR)thisHwnd;
	copydata.lpData = (LPVOID)sendData;
	copydata.cbData = size;
	return ::SendMessage( targetHwnd, WM_COPYDATA, type, (LPARAM)&copydata );
}

IPCCopyDataHandle *IPCCopyDataServer::GetHandle(HWND hWnd, LPCTSTR name,IIPCReceiveCallback* callback )
{
	IPCCopyDataHandle *handle = SearchHandle(hWnd);
	if( NULL != handle )
		return handle;

	handle = CreateNewHandle(hWnd, name);
	handle->SetReadCallback(callback);
	SendIPCCopyDataPacket( hWnd, m_msgWnd->m_hWnd, IPC_COPYDATA_PACKET_CONNECT, 
							(LPVOID)m_hostkey, lstrlen(m_hostkey) * 2);
	return handle;
}

IPCCopyDataHandle *IPCCopyDataServer::CreateNewHandle(HWND hWnd, LPCTSTR name)
{
	IPCCopyDataHandle *handle = new IPCCopyDataHandle(hWnd, m_msgWnd->m_hWnd, name, this);
	if( NULL == handle )
		return NULL;

	m_handleMap.insert( std::make_pair(hWnd, handle) );
	return handle;
}

int IPCCopyDataServer::Receive( IPC_COPYDATA_COMMUNICATION_TYPE type, COPYDATASTRUCT *data )
{
	switch( type )
	{
	case IPC_COPYDATA_PACKET_CONNECT:
		{
		ProcessConnect(data);	
		}
		break;
	case IPC_COPYDATA_PACKET_CLOSE:
		{
		ProcessClose(data);
		}
		break;
	case IPC_COPYDATA_PACKET_RECEIVE:
		{
		ProcessReceive(data);
		}
		break;
	case IPC_COPYDATA_PACKET_START:
		{
		ProcessReceiveStart(data);
		}
		break;
	default:
		break;
	}
	return 1;
}

void IPCCopyDataServer::SetHooker( IPCCopyDataServerEventHooker *hooker )
{
	m_hooker = hooker;
}

bool IPCCopyDataServer::ProcessConnect( COPYDATASTRUCT *data )
{
	IPCCopyDataHandle *handle = SearchHandle((HWND)data->dwData);
	if( NULL != handle )
		return false;

	int name_size = data->cbData;
	if( name_size < 0 )
		return false;

	TCHAR name[IPC_COPYDATA_MAX_SEND_SIZE];
	memcpy( name, data->lpData, name_size );
	name[name_size/2] = 0;
	handle = CreateNewHandle((HWND)data->dwData, name);
	if( NULL == handle )
		return false;

	if( m_hooker )
	{
		m_hooker->Connected(handle);
	}

	return true;
}

bool IPCCopyDataServer::ProcessClose( COPYDATASTRUCT *data )
{
	IPCCopyDataHANDLE_MAP::iterator it = m_handleMap.find( (HWND)data->dwData );
	if( it == m_handleMap.end() )
	{
		return false;
	}

	if( m_hooker )
	{
		m_hooker->Close((*it).second);
	}

	IPCCopyDataHandle *handle = (*it).second;
	handle->CloseWithoutSend();
	m_handleMap.erase(it);
	delete handle;
	return true;
}

bool IPCCopyDataServer::ProcessReceive( COPYDATASTRUCT *data )
{
	IPCCopyDataHandle *handle = SearchHandle((HWND)data->dwData);
	if( NULL == handle )
		return false;

	handle->ReceiveData((const BYTE *)data->lpData, data->cbData);
	return true;
}

IPCCopyDataHandle *IPCCopyDataServer::SearchHandle(HWND hWnd)
{
	IPCCopyDataHANDLE_MAP::iterator it = m_handleMap.find( hWnd );
	if( it == m_handleMap.end() )
	{
		return NULL;
	}

	return (*it).second;
}

bool IPCCopyDataServer::ProcessReceiveStart( COPYDATASTRUCT *data )
{
	HWND hWnd = (HWND)data->dwData;
	IPCCopyDataHandle *handle = NULL;
	IPCCopyDataHANDLE_MAP::iterator it = m_handleMap.find( hWnd );
	if( it == m_handleMap.end() )
	{
		return false;
	}

	handle = (*it).second;

	int size;
	memcpy( &size, data->lpData, sizeof(int) );
	handle->SetContinueMode(size);
	return true;
}

class MakeHandleListFunctor
{
public:
	void operator()( std::pair<HWND, IIPCHandle *> data  )
	{
		m_handleList->push_back(data.second);
	}

	void Set( IIPCHandleList *list )
	{
		m_handleList = list;
	}

private:
	IIPCHandleList *m_handleList;
};

IIPCHandleList IPCCopyDataServer::GetHandleList()
{
	IIPCHandleList handleList;

	MakeHandleListFunctor functor;
	functor.Set( &handleList );
	std::for_each( m_handleMap.begin(), m_handleMap.end(), functor );

	return handleList;
}

IPCCopyDataHandle::IPCCopyDataHandle(HWND hWnd, HWND serverWnd, LPCTSTR name, IPCCopyDataServer* server) 
                                     : m_hWnd(hWnd), m_serverWnd(serverWnd), m_key(name),m_keepSize(0),
									 m_continueMode(false), m_continueSize(0), m_buffer(NULL), m_server(server), IIPCHandle()
{
}

IPCCopyDataHandle::~IPCCopyDataHandle()
{
}

bool IPCCopyDataHandle::Read(IPCPacket& packet)
{
	if( true == m_queue.empty() )
		return false;

	packet = *(m_queue.begin());
	m_queue.pop_front();
	return true;
}

bool IPCCopyDataHandle::IsValid()
{
	TCHAR *name = (TCHAR *)m_key.GetData();
	return IsHandleExist( name );
}

bool IPCCopyDataHandle::Write(const BYTE *data, int size)
{
	if( NULL == m_hWnd || FALSE == IsValid() )
		return false;

	BYTE sendbuffer[IPC_COPYDATA_MAX_SEND_SIZE] = {0,};

	if( size <= IPC_COPYDATA_MAX_SEND_SIZE )
	{
		memcpy( sendbuffer, data, size );
		SendIPCCopyDataPacket( m_hWnd, m_serverWnd, IPCCopyDataServer::IPC_COPYDATA_PACKET_RECEIVE,
								(LPVOID)sendbuffer, size );
	}
	else
	{
		BYTE *ptr = (BYTE *)data;

		SendIPCCopyDataPacket( m_hWnd, m_serverWnd, IPCCopyDataServer::IPC_COPYDATA_PACKET_START,
			(LPVOID)&size, sizeof(int) );

		while( size > 0 )
		{
			int copy_size = min(size, IPC_COPYDATA_MAX_SEND_SIZE);
			memcpy( sendbuffer, ptr, copy_size );
			SendIPCCopyDataPacket( m_hWnd, m_serverWnd, IPCCopyDataServer::IPC_COPYDATA_PACKET_RECEIVE,
				(LPVOID)sendbuffer, copy_size );

			size -= copy_size;
			ptr += copy_size;
		}
	}

	return true;
}

IIPCKey *IPCCopyDataHandle::GetKey()
{
	return &m_key;
}

void DeletePacket( IPCPacket packet );

void IPCCopyDataHandle::Close()
{
	if( NULL == m_hWnd )
		return;

	m_callback = NULL;

	IPCPacketQueue::iterator it;
	for( it = m_queue.begin(); it != m_queue.end(); it++ )
	{
		IPCPacket packet = (*it);
		packet.Release();
	}

	m_hWnd = NULL;
}

void IPCCopyDataHandle::CloseWithoutSend()
{
	Close();
}

void IPCCopyDataHandle::CloseWithSend()
{
	HWND tempHwnd = m_hWnd;
	Close();
	SendIPCCopyDataPacket( m_hWnd, m_serverWnd, IPCCopyDataServer::IPC_COPYDATA_PACKET_CLOSE,
		(LPVOID)NULL, 0 );
}

void IPCCopyDataHandle::PushPacketInQueue(IPCPacket& packet)
{
	if( m_callback )
	{
		m_callback->Callback( packet, this );
	}
	else
	{
		m_queue.push_back(packet);
	}
}

void IPCCopyDataHandle::ReceiveData( const BYTE *data, int size )
{
	if( m_continueMode )
	{
		BYTE *ptr = m_buffer + m_keepSize;
		memcpy( ptr, data, size);

		m_keepSize += size;
		if( m_keepSize == m_continueSize )
		{
			IPCPacket packet;
			packet.size = m_continueSize;
			packet.data = m_buffer;
			PushPacketInQueue(packet);
			m_buffer = NULL;
			m_keepSize = 0;
			m_continueSize = 0;
			m_continueMode = false;
		}
	}
	else
	{
		IPCPacket packet;
		packet.size = size;
		packet.data = new BYTE[size];
		if( NULL == packet.data )
			return;

		memcpy( packet.data, data, size );
		PushPacketInQueue(packet);
	}
}

void IPCCopyDataHandle::SetContinueMode( int size )
{
	if( size < IPC_COPYDATA_MAX_SEND_SIZE )
		return;

	m_continueMode = true;
	m_continueSize = size;
	m_keepSize = 0;

	if( NULL != m_buffer )
	{
		delete m_buffer;
	}

	m_buffer = new BYTE[size];
}

};